<?php

namespace App\Http\Services\V4;

use App\Http\Requests\V3\InvoiceRequest;
use App\Models\{Badrshop, BillAdd, Client, ClientTransaction, Invoice, InvoiceDetails, InvoiceDetailVat, Item, Log, SalePoint, StoreItem};
use Illuminate\Support\Facades\{DB, Validator};
use Carbon\Carbon;
use Exception;
use Illuminate\Support\Arr;
use PDOException;

class InvoiceService
{
    protected Badrshop $shop;
    protected SalePoint $salePoint;
    protected SalePoint $networkSalePoint;
    protected Client $client;
    protected Invoice $invoice;
    protected $type;
    protected array $details = [];
    protected array $errors  = [];
    protected array $saved_invoices = [];
    protected array $points = [];

    public function __construct()
    {
        logBill("URL : " . request()->getUri());
        logBill("Start Store Bills  | User Id : " . authId());
        $this->shop = Badrshop::select('serial_id', 'company_ratio', 'add_delivery_fees_to_the_cash_drawer', 'date_type')->first();
    }

    public function handler(array $data = [], int $type = 1)
    {
        logBill(json_encode($data));
        logBill("The invoice type : $type");

        try {
            DB::beginTransaction();
            $this->type = $type;
            foreach ($data['bills'] as $bill) {
                $bill['all_payment'] = $bill['payment'] + $bill['network_paid'];
                $result = Validator::make($bill, (new InvoiceRequest)->single_rules());
                if ($result->fails()) {
                    $this->errors[$bill['local_bill_no']] = Arr::flatten( $result->errors()->getMessages() );
                    logBill(json_encode($result->errors()), 'warning');
                    continue;
                }

                $this->resetProperties($bill); // make all properties null

                try {
                    // To create new invoice
                    $this->createInvoice($bill);
                } catch(\Throwable $th) {
                    $this->handleDuplicateInvoices($bill, $th);
                    continue;
                }

                // To update the client balance [increment or decrement]
                $this->updateClientBalance($bill['total'], $bill['all_payment']);

                foreach ($this->points as $key => $salePoint) {
                    // To update the sale poind money_point [increment or decrement]
                    $this->updateSalePointBalance($salePoint, $bill[$key]);

                    // To create client transaction
                    $this->createTransaction($salePoint, $bill[$key], $bill);
                }

                // To create bill details
                $this->itemsProcess($bill['sale_details']);

                // To create invoice log
                $this->newInvoiceLog();

                $this->saved_invoices[] = ['bill_no' => $this->invoice->bill_no, 'bill_id' => $this->invoice->id];
            }
            DB::commit();

            if ( count( $this->saved_invoices )  == 0) throw new Exception("لم يتم حفظ اي فواتير");

            $response = ['status' => 200, 'data' => $this->saved_invoices];
            logBill("End Success Store Bills  | User Id : " . authId());
        } catch (Exception $e) {
            logBill("Exception Object : " . json_encode($e), 'debug');
            logBill("End Failed Store Bills  | User Id : " . authId(), 'debug');
            $response = ['status' => 500, 'message' => $e->getMessage(), 'errors' => $this->errors];
        }

        logBill("response : " . json_encode($response));
        logBill( str_repeat('_', 100), 'notice');
        return $response;
    }

    protected function handleDuplicateInvoices(array $bill, Exception $th)
    {
        logBill(json_encode($th), 'debug');
        if($th instanceof PDOException) {
            $row = Invoice::where([
                    'client_id'    => $this->client->id,
                    'net_price'    => $bill['total'],
                    'invoice_date' => Carbon::parse($bill['date_time'])
                ])->select(['id', 'bill_no'])->first();

            if ($row) {
                $this->saved_invoices[] = ['bill_no' => $row->bill_no, 'bill_id' => $row->id];
            }
        }
    }

    protected function updateClientBalance(float $total = 0, float $payment = 0): void
    {
        if ($this->type == Invoice::SALES)
            $this->client->increment('balance', $total - $payment);
        else
            $this->client->decrement('balance', $total - $payment);
    }

    protected function updateSalePointBalance(SalePoint $salePoint, float $payment): void
    {
        if ($this->type == Invoice::SALES) {
            $salePoint->increment('money_point', $payment);
        } else {
            $salePoint->decrement('money_point', $payment);
        }
    }

    protected function createInvoice(array $bill): void
    {
        $this->invoice = Invoice::create([
            'invoice_date'              => Carbon::parse($bill['date_time']),
            'local_bill_no'             => $bill['local_bill_no'],
            'store_id'                  => $this->salePoint->store_id,
            'currency_id'               => $this->salePoint->currency_id,
            'currency_value'            => $this->salePoint->currency->value,
            'invoice_type'              => $this->type,
            'bill_total'                => $bill['net'],
            'net_price'                 => $bill['total'],
            'payment'                   => $bill['all_payment'],
            'bill_payment'              => $bill['all_payment'],
            'rest'                      => $bill['total'] - $bill['all_payment'],
            'balance'                   => $this->client->balance,
            'client_id'                 => $bill['client_id'],
            'invoice_total_disc_status' => $this->shop->company_ratio == 2 ? 0 : 1,
            'insatllments_add'          => $bill['installment_add_precent'] ?? 0,
            'supplier_invoice_no'       => $bill['supplier_invoice_no'] ?? null,
            'reference_id'              => $bill['reference_id'] ?? 0,
            'fee'                       => $bill['delivery_fees'] ?? 0,
            'fee_status'                => $this->shop->add_delivery_fees_to_the_cash_drawer,
            'invoice_total_disc'        => $bill['discountCashValue'],
        ]);

        $this->invoice->discount_percentage = $this->invoice->bill_total ? ($this->invoice->invoice_total_disc / $this->invoice->bill_total) * 100 : 0;
    }

    protected function setDiscountPercentage()
    {
        if ($this->invoice->bill_total > 0) {
            $this->invoice->discount_percentage = ($this->invoice->invoice_total_disc / $this->invoice->bill_total) * 100;
        } else {
            $this->invoice->discount_percentage = 0;
        }
    }

    protected function itemsProcess(array $items): void
    {
        $arr_vats = [];
        foreach ($items as $item) {
            $item_obj = Item::with(['additions' => fn($query) => $query->where('vat_use', '!=', 2) ])->find($item['item_id']);
            $item['real_quantity'] = $item['unitValue'] * $item['quantity'];
            $item['price_without_discount'] = $item['price'] - ($item['price'] * ($this->invoice->discount_percentage / 100));
            $item['quantity_after'] = $this->updateUnitQuantity($item);
            $this->createInvoiceDetails($item, $item_obj);
            $item['object'] = $item_obj;
            $addition = $item_obj->additions->first();
            if ($item_obj->vat_state !== 0 && $addition) {
                $this->createInvoiceItemVat($addition, $item, $arr_vats);
            }
        }

        if (count($arr_vats)) InvoiceDetailVat::insert($arr_vats);
        InvoiceDetails::insert($this->details);
    }

    protected function updateUnitQuantity(array $item): float
    {
        $query = StoreItem::where(['item_id' => $item['item_id'], 'store_id' => $this->salePoint->store_id]);
        if ($this->type == Invoice::SALES)
            $query->decrement('store_quant', $item['real_quantity']);
        else
            $query->increment('store_quant', $item['real_quantity']);

        return $query->first()->store_quant;
    }

    protected function createInvoiceDetails(array $item, Item $row): void
    {
        $item['price_without_vat'] = $item['price_without_discount'];
        if ($this->shop->apply_vats && $row->vat_state == 2) {
            $item['price_without_vat'] = $this->caclAmount($row);
        }

        $this->details[] = [
            'shop_id'          => shopId(),
            'invoice_id'       => $this->invoice->id,
            'safe_point_id'    => $this->salePoint->id,
            'store_id'         => $this->salePoint->store_id,
            'add_user'         => authId(),
            'item_id'          => $item['item_id'],
            'quantity'         => $item['quantity'],
            'quantity_after'   => $item['quantity_after'],
            'unit_id'          => $item['unit_id'],
            'net_price'        => 0.00,
            'pay_price'        => $row->pay_price * $item['unitValue'],
            'total_price'      => $item['price'] * $item['quantity'],
            'sale_price'       => $item['price'],
            'basic_sale_price' => $item['price'],
            'basic_pay_price'  => $row->pay_price * $item['unitValue'],
            'av_pay_price'     => $row->pay_price,
            'ac_pay_price'     => $row->pay_price * $item['unitValue'],
            'item_about'       => '',
            'discount'         => $this->invoice->discount_percentage,
            'discount_type'    => $this->invoice->invoice_total_disc_type,
            'item_price_with_discount' => $item['price_without_discount'],
            'price_without_vat'=> $item['price_without_vat'],
        ];
    }

    protected function caclAmount($item)
    {
        $vat_use = in_array($this->type, [0, 2]) ? [0, 2] : [0, 1];
        $adds = BillAdd::select('addition_value', 'check_addition', 'id')
                        ->where(['shop_id' => auth()->user()->shop_id, 'add_role' => false])
                        ->whereHas('itemVats', function($query) use($item, $vat_use) {
                            $query->where('item_id', $item->id)->whereIn('vat_use', $vat_use);
                        })->get();

        $vat_value = $this->calcVats($adds, $item->price);
        return $item->price - $vat_value;
    }

    protected function calcVats($adds, $price)
    {
        $vat_value = 0;
        foreach ($adds as $add) {
            if ( $add->check_addition ) {
                $vat_value += $add->addition_value;
            } else {
                $vat_value += $price - ($price / ( 1 + ($add->addition_value / 100) ));
            }
        }

        return $vat_value;
    }

    protected function createInvoiceItemVat(BillAdd $addition, array $item_detail, array &$vats): void
    {
        $vats[] = [
            'invoice_id'    => $this->invoice->id,
            'item_id'       => $item_detail['item_id'],
            'unit_id'       => $item_detail['unit_id'],
            'vat_id'        => $addition->id,
            'vat_type'      => $addition->check_addition,
            'vat_value'     => $addition->addition_value,
            'vat_operation' => $addition->vat_operation,
            'vat_weight'    => $this->calculateVat($item_detail, $addition),
            'shop_id'       => shopId()
        ];
    }

    protected function calculateVat(array $item, BillAdd $row)
    {
        if ($row->check_addition)
            return $row->addition_value;

        $price = $item['price_without_discount'] * $item['quantity'];

        if ($item['object']->vat_state == 1) {
            return $price * ($row->addition_value / 100);
        } else {
            $percentage = ($row->addition_value / 100) + 1;
            return $price - ($price / $percentage );
        }
    }

    protected function createTransaction(SalePoint $salePoint, float $payment, array $bill): void
    {
        try {
            ClientTransaction::create([
                'user_id'           => authId(),
                'pay_day'           => $this->invoice->invoice_date,
                'notes'             => $this->invoice->notes,
                'amount'            => $payment,
                'bill_net_total'    => $this->invoice->net_price,
                'bill_id'           => $this->invoice->id,
                'date_time'         => day_now($this->shop->date_type),
                'balance'           => $this->client->balance,
                'client_id'         => $this->client->id,
                'safe_point_id'     => $salePoint->id,
                'safe_balance'      => $salePoint->money_point,
                'effect'            => $this->type == Invoice::SALES,
                'type'              => $this->type == Invoice::SALES,
                'sale_back_id'      => 0,
                'problem_id'        => 0,
                'supplier_id'       => 0,
                'salary_month'      => 0,
                'OutTransactionID'  => 0,
                'spend_id'          => 0,
            ]);
        } catch(\Throwable $th) {
            $this->handleDuplicateInvoices($bill, $th);
        }
    }

    protected function newInvoiceLog(): void
    {
        $message = "تم انشاء فاتورة " . $this->getInvoiceTypeName();
        Log::create(['event' => $message . trans('log.invoice_no') . $this->invoice->bill_no, 'bill_id' => $this->invoice->id, 'type' => $this->type ?? 1 ]);
    }

    protected function getInvoiceTypeName(): string
    {
        $name = '';
        switch ($this->type) {
            case 0 :
                $name = "مشتريات";
                break;
            case 1 :
                $name = "مبيعات";
                break;
            case 2 :
                $name = "مرتجع مشتريات";
                break;
            case 3 :
                $name = "مرتجع مبيعات";
                break;
        }

        return $name;
    }

    protected function resetProperties(array $bill): void
    {
        $this->salePoint        = SalePoint::select('store_id', 'id', 'currency_id', 'money_point')->whereId($bill['sale_point'])->with('currency')->first() ?? new SalePoint();
        $this->networkSalePoint = SalePoint::select('store_id', 'id', 'currency_id', 'money_point')->where('point_type', 2)->whereId($bill['network_id'])->first() ?? new SalePoint();
        $this->client           = Client::select('id', 'balance')->find($bill['client_id']) ?? new Client();
        $this->invoice          = new Invoice();
        $this->details          = [];

        $this->points['payment'] = $this->salePoint;
        if ($bill['network_paid'] > 0) {
            $this->points['network_paid'] = $this->networkSalePoint;
        }
    }
}
