<?php

namespace App\Http\Services\V4;

use App\Http\Requests\V4\NewRequestRequest;
use App\Models\{Badrshop, Client, Item, Log, NewRequest, NewRequestDetail, SalePoint, Store};
use Illuminate\Support\Facades\{DB, Validator};
use Carbon\Carbon;
use Exception;
use PDOException;

class NewRequestService
{
    protected Badrshop $shop;
    protected SalePoint $salePoint;
    protected Store $store;
    protected Client $client;
    protected NewRequest $request;
    protected Calculator $calculator;
    protected $type;
    protected $items;
    protected array $details = [];
    protected array $saved_requests = [];

    public function __construct()
    {
        $this->shop = Badrshop::select('serial_id', 'company_ratio', 'add_delivery_fees_to_the_cash_drawer', 'date_type')->first();
    }

    public function handler(array $data = [], int $type = 1)
    {
        $this->type = $type;
        logger(json_encode($data));
        try {
            DB::beginTransaction();
                foreach ($data['requests'] as $request) {
                    $result = Validator::make($request, (new NewRequestRequest())->single_rules());
                    if ($result->fails()) {
                        logger(json_encode($result->errors()));
                        continue;
                    }

                    if ($id = $this->checkRequestExists($request)) {
                        $this->saved_requests[] = ['receipt_number' => $request['receipt_number'], 'bill_id' => $id];
                        continue;
                    }

                    $this->resetProperties(); // make all properties null
                    $this->salePoint = SalePoint::select('id', 'currency_id', 'money_point')->whereId( salePointId() )->first();
                    $this->store     = Store::select('id')->whereId( $request['store_id'] )->first();
                    $this->client    = Client::select('id', 'balance', 'price')->find($request['client_id']);
                    $this->items     = collect( $request['items'] );
                    $this->setCalculator($request['items'][0]);

                    try {
                        // To create new invoice
                        $this->createRequest($request);
                    } catch(\Throwable $th) {
                        $this->handleDuplicateRequest($request['date_time'], $th);
                        continue;
                    }

                    $this->itemsProcess();

                    // To create invoice log
                    $this->newInvoiceLog();
                    $this->saved_requests[] = ['receipt_number' => $this->request->no_bill, 'bill_id' => $this->request->id];
                }
            DB::commit();

            return ['status' => 200, 'data' => $this->saved_requests];
        } catch (Exception $e) {
            return ['status' => 500, 'message' => $e->getMessage()];
        }
    }

    protected function handleDuplicateRequest(string $date, Exception $th)
    {
        if ($th instanceof PDOException) {
            $row = NewRequest::where(['client_id' => $this->client->id, 'net' => $this->calculator->net_price, 'invoice_date' => Carbon::parse($date)])
                                ->select(['id',  'no_bill'])->first();
            if ($row) {
                $this->saved_requests[] = ['receipt_number' => $row->no_bill, 'bill_id' => $row->id];
            }
        }
    }

    protected function createRequest(array $request): void
    {
        $this->request = NewRequest::create([
            'client_id'    => $this->client->id,
            'store_id'     => $this->store->id,
            'total'        => $this->calculator->total,
            'net'          => $this->calculator->net_price,
            'no_bill'      => $request['receipt_number'],
            'notes'        => $request['notes'] ?? '',
            'invoice_date' => Carbon::parse($request['date_time']),
        ]);
    }

    protected function itemsProcess(): void
    {
        foreach ($this->items as $item) {
            $this->createInvoiceDetails($item);
        }

        NewRequestDetail::insert($this->details);
    }

    protected function createInvoiceDetails(array $item): void
    {
        $this->details [] = [
                    'shop_id'          => shopId(),
                    'request_id'       => $this->request->id,
                    'item_id'          => $item['item_id'],
                    'quantity'         => $this->calculator->quantity,
                    'price'            => $this->calculator->total,
                    'sale_price'       => $this->calculator->price,
                    'unit_id'          => Item::select('unit_id')->find($item['item_id'])->unit_id ?? 0,
                    'water'            => $this->calculator->water,
                    'protein'          => $this->calculator->protein,
                    'fats'             => $this->calculator->fats,
                    'water_discount'   => $this->calculator->water_discount,
                    'protein_bonus'    => $this->calculator->protein_bonus,
                    'fats_bonus'       => $this->calculator->fats_bonus,
                ];
    }

    protected function checkRequestExists(array $request)
    {
        return NewRequest::select('id')->where('created_at', Carbon::parse($request['date_time']))
                            ->where('store_id', $request['store_id'])->where('client_id', $request['client_id'])
                            ->where('no_bill', $request['receipt_number'])->first()->id ?? false;
    }

    protected function newInvoiceLog()
    {
        Log::create(['event' => trans('log.new_sale_invoice') . trans('log.invoice_no') . $this->request->id, 'type' => $this->type]);
    }

    protected function setCalculator(array $item)
    {
        $this->calculator = new Calculator();

        $this->calculator->setQuantity($item['quantity'])->setWater($item['water'])
                        ->setProtein($item['protein'])->setFats($item['fats'])
                        ->setPrice($this->client->price)->calculate();
    }

    protected function resetProperties()
    {
        $this->salePoint   = new SalePoint();
        $this->store       = new Store();
        $this->client      = new Client();
        $this->request     = new NewRequest();
        $this->calculator  = new Calculator();
        $this->details     = [];
        $this->items       = collect();
    }
}
